﻿
using MeshVR;
using MVR.FileManagementSecure;
using SimpleJSON;
using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Security.Cryptography;
using System.Text;
using System.Text.RegularExpressions;
using UnityEngine;
using UnityEngine.UI;



namespace Vam_preset_maker
{

    public class PresetMaker : MVRScript
    {
        UIDynamicButton processVarButton;
        UIDynamicButton processLooksButton;
        UIDynamicButton processSaves;
        UIDynamicButton retakeScreenShots;

        UIDynamicToggle SaveAppearance;
        UIDynamicToggle SaveClothing;
        UIDynamicToggle SaveSkin;
        UIDynamicToggle SaveHair;

        JSONStorableBool SaveAppearanceBool = new JSONStorableBool("AppearancePresets", true);
        JSONStorableBool SaveClothingBool = new JSONStorableBool("ClothingPresets", true);
        JSONStorableBool SaveSkinBool = new JSONStorableBool("SkinPresets", true);
        JSONStorableBool SaveHairBool = new JSONStorableBool("HairPresets", true);

        List<JSONStorableBool> _saveOptions;
        List<string> CreatorsList = new List<string>() { "All" };
        Dictionary<string, List<string>> CreatorPaths = new Dictionary<string, List<string>>();

        DAZCharacterSelector _dazCharacterSelector;

        CamPositions camPositions = new CamPositions();

        bool inScanExistingVapFiles = false;
        bool inStoreAuthors = false;
        bool waitingOnScreenShot = false;
        bool needToSaveCams = true;
        bool IsProcessing = false;
        bool StopRequested = false;
        int fileCount = 0;
        int lookCount = 0;

        CacheSystem cacheSystem;

        FilePaths filePaths;

        JSONStorableStringChooser jssc;

        Transform _headTransform;
        
        InfoBox infoBox;

        public override void Init()
        {
            //create folders
            if (!FileManagerSecure.DirectoryExists("Custom/PluginData/Chokaphi/PresetMaker/"))
            {
                FileManagerSecure.CreateDirectory("Custom/PluginData/Chokaphi/PresetMaker/");
            }

            if (base.containingAtom.mainController == null)
            {
                SuperController.LogError("Please add this plugin to a PERSON atom.");
                return;
            }

            _headTransform = containingAtom.rigidbodies.Where(x => x.name == "head").FirstOrDefault().transform;
            _dazCharacterSelector = containingAtom.GetComponentInChildren<DAZCharacterSelector>();

            camPositions.StoreCamPosition("VAMCAM");
            camPositions.StoreCamPosition("HeadPosition", _headTransform);

            cacheSystem = new CacheSystem(this);

            JSONStorableString uiInfoBox = new JSONStorableString("uiSetupTimer", "");
            infoBox = new InfoBox(uiInfoBox);

            UIDynamicTextField t = CreateTextField(uiInfoBox);
            RectTransform rt =  t.backgroundImage.GetComponent<RectTransform>();
            Vector2 v = rt.sizeDelta;
            rt.sizeDelta = new Vector2(500, v.y);
            rt.transform.localPosition += new Vector3(250, 0, 0);
            UIDynamic spacerRight = CreateSpacer(true);
            spacerRight.height = 215;
  
            StartCoroutine(ScanSetup());

        }

        void BrowserCallBack(string path)
        {
            infoBox.Status = "Processing Save Files";
            StartCoroutine(ProcessLooks(path));
        }

        void EnableUIButtons(bool state)
        {
            processVarButton.button.enabled = state;
            processLooksButton.button.enabled = state;
            processSaves.button.enabled = state;
            retakeScreenShots.button.enabled = state;
        }

        void CreateUI()
        {
            SaveAppearance = CreateToggle(SaveAppearanceBool);
            SaveClothing = CreateToggle(SaveClothingBool);
            SaveSkin = CreateToggle(SaveSkinBool);
            SaveHair = CreateToggle(SaveHairBool);
            _saveOptions = new List<JSONStorableBool>() { SaveAppearanceBool, SaveClothingBool, SaveHairBool, SaveSkinBool };

            jssc = new JSONStorableStringChooser("Creators", CreatorsList, "All", "Select Creator to filter by");
            UIDynamicPopup pop = CreateScrollablePopup(jssc, true);

            processVarButton = CreateButton("Process Var Folder", true);
            processVarButton.button.onClick.AddListener(() =>
            {
                infoBox.Status = "Processing VAR Files";
                StartCoroutine(ProcessLooks("AddonPackages"));
            });

            processLooksButton = CreateButton("WIP Feature - Process Looks", true);
            processLooksButton.button.onClick.AddListener(() =>
            {
                infoBox.Status = "Processing Looks folder";
                StartCoroutine(ProcessLooks("Saves/Person/appearance/"));
            });

            processSaves = CreateButton("WIP Feature - Scan Saved Scenes", true);
            processSaves.button.onClick.AddListener(() =>
            {
                infoBox.Status = "Processing Saved Scenes";
                SuperController.singleton.GetDirectoryPathDialog(BrowserCallBack, "Saves/Scenes");
            });

            retakeScreenShots = CreateButton("Retake Camera Positions");
            retakeScreenShots.button.onClick.AddListener(() =>
            {
                StartCoroutine(StoreCameraPostions(true));
            });

            infoBox.Message = "UI Created. Ready to start Processing";

            #region testbuttons
            //UIDynamicButton Cam1 = CreateButton("check ss");
            //Cam1.button.onClick.AddListener(() =>
            //{
            //    SuperController.LogError("Height is " + SuperController.singleton.playerHeightAdjust);
            //    if (camPositions.LoadCamPosition("HeadPosition") == null)
            //    {
            //        SuperController.LogError("head pos null");
            //    }

            //});

            //UIDynamicButton Cam3 = CreateButton("Sload cam");
            //Cam3.button.onClick.AddListener(() =>
            //{
            //    camPositions.UpdateHeadDeltas(HeadTransform);
            //    camPositions.LoadCamPosition("AppearancePresets");

            //});
            //UIDynamicButton Cam4 = CreateButton("reset to start pos");
            //Cam4.button.onClick.AddListener(() =>
            //{

            //});
            //UIDynamicButton Cam5 = CreateButton("Test 2");
            //Cam5.button.onClick.AddListener(() =>
            //{
            //    Transform monitorCenterCam = SuperController.singleton.MonitorCenterCamera.transform;
            //    Transform camRig = SuperController.singleton.GetAtomByUid("[CameraRig]").transform;
            //    Transform navRig = SuperController.singleton.navigationRig.transform;

            //    Transform headController = containingAtom.freeControllers.Where(x => x.name == "headControl").FirstOrDefault().transform;

            //    string testString = "Mon cam {0, -15} {1, -15} {2, -15}\nRig cam {3, -15} {4, -15} {5, -15}\nhead c {6, -15} {7, -15} {8, -15}";


            //    SuperController.LogError(string.Format(testString,  monitorCenterCam.localPosition.x, monitorCenterCam.localPosition.y, monitorCenterCam.localPosition.z, navRig.localPosition.x, navRig.localPosition.y , navRig.localPosition.z, headController.localPosition.x, headController.localPosition.y, headController.localPosition.z) + "\n" + SuperController.singleton.playerHeightAdjust);
            //    //SuperController.LogError(string.Format(testString, monitorCenterCam.localEulerAngles.x, monitorCenterCam.localEulerAngles.y, monitorCenterCam.localEulerAngles.z, navRig.localEulerAngles.x, navRig.localEulerAngles.y, navRig.localEulerAngles.z, navRig.localEulerAngles.x, navRig.localEulerAngles.y, navRig.localEulerAngles.z));
            //    SuperController.LogError(headController.position.x + " "+ headController.position.y + " "+  headController.position.z);


            //});

            #endregion
        }


        IEnumerator ScanSetup()
        {
            StartCoroutine(ScanExistingVapFiles());
            StartCoroutine(StoreAuthors());

            yield return new WaitWhile(() => inScanExistingVapFiles || inStoreAuthors);
            
            CreateUI();
            infoBox.Status = "UI created";

            yield return null;
        }

        IEnumerator ScanExistingVapFiles()
        {
            inScanExistingVapFiles = true;
            int count = 0;
            foreach (string filePath in GetFiles("Custom/Atom/Person/Appearance/", "*.vap"))
            {
                string file = FileManagerSecure.ReadAllText(filePath);
                yield return new WaitForEndOfFrame();
                string hash = cacheSystem.ComputeNodeHash(file);
                yield return new WaitForEndOfFrame();
                cacheSystem.InNodeCache(hash, filePath);
                count++;
                infoBox.VapScan = "Scanning Appearance";
                infoBox.VapCount = count;
            }
      
            foreach (string filePath in GetFiles("Custom/Atom/Person/Clothing/", "*.vap"))
            {
                string file = FileManagerSecure.ReadAllText(filePath);
                yield return new WaitForEndOfFrame();
                string hash = cacheSystem.ComputeNodeHash(file);
                yield return new WaitForEndOfFrame();
                cacheSystem.InNodeCache(hash, filePath);
                count++;
                infoBox.VapScan = "Scanning Clothing";
                infoBox.VapCount = count;
            }

            foreach (string filePath in GetFiles("Custom/Atom/Person/Hair/", "*.vap"))
            {
                string file = FileManagerSecure.ReadAllText(filePath);
                yield return new WaitForEndOfFrame();
                string hash = cacheSystem.ComputeNodeHash(file);
                yield return new WaitForEndOfFrame();
                cacheSystem.InNodeCache(hash, filePath);
                count++;
                infoBox.VapScan = "Scanning Hair";
                infoBox.VapCount = count;
            }

            foreach (string filePath in GetFiles("Custom/Atom/Person/Skin/", "*.vap"))
            {
                string file = FileManagerSecure.ReadAllText(filePath);
                yield return new WaitForEndOfFrame();
                string hash = cacheSystem.ComputeNodeHash(file);
                yield return new WaitForEndOfFrame();
                cacheSystem.InNodeCache(hash, filePath);
                count++;
                infoBox.VapScan = "Scanning Skin";
                infoBox.VapCount = count;
            }
            cacheSystem.SaveCache();
            inScanExistingVapFiles = false;
            yield return null;
        }

        IEnumerator StoreAuthors()
        {
            inStoreAuthors = true;
            int count = 0;
            foreach (string filePath in GetFiles("AddonPackages", "*.var"))
            {
                count++;
                //uiInfoBox.SetVal("Creating the Author filter, UI will be created once finished. \nVar's found " + count + " Authors found = " + CreatorsList.Count);
                infoBox.Status = "Creating the Author filter";
                infoBox.Message = "UI will be created once scans are finished";
                infoBox.VarCount = count;
                infoBox.AuthorCount = CreatorsList.Count;

                yield return new WaitForEndOfFrame();
                string varName = FileManagerSecure.GetFileName(filePath);

                string[] varNames = varName.Split('.');
                if (!CreatorsList.Contains(varNames[0]))
                {
                    CreatorPaths[varNames[0]] = new List<string>();
                    CreatorsList.Add(varNames[0]);
                    yield return new WaitForEndOfFrame();
                }

                CreatorPaths[varNames[0]].Add(filePath);
            }

            CreatorsList.Sort();
            infoBox.AuthorCount = CreatorsList.Count;
            inStoreAuthors = false;

            yield return null;
        }

        IEnumerator StoreCameraPostions(bool force = false)
        {
            needToSaveCams = false;
            Text help = SuperController.singleton.screenshotPreview.GetComponentsInChildren<Text>().Where(x => x.name == "HelpText").FirstOrDefault();
            camPositions.OrigHelpText = help.text;
            camPositions.OrigHelpColor = help.color;

            //use saved cam from scene or do we use a set new positions
            if (camPositions.LoadCamPosition("VAMCAM") == null || force)
            {
                camPositions.StoreCamPosition("VAMCAM");
            }
            //record cam position vs head position
            if (camPositions.LoadCamPosition("HeadPosition") == null || force)
            {
                camPositions.StoreCamPosition("HeadPosition", _headTransform);
            }

            if (camPositions.LoadCamPosition("AppearancePresets") == null || force)
            {
                camPositions.StoreCamPosition("AppearancePresets");
                waitingOnScreenShot = true;
                SuperController.singleton.DoSaveScreenshot("Custom/PluginData/Chokaphi/PresetMaker/AppearancePresets", RecordCamPos);
                help.text = "Set Camera you want to use for Appearance Preset";
                help.color = Color.cyan;

                yield return new WaitWhile(() => waitingOnScreenShot);
                yield return new WaitForSecondsRealtime(1);

            }

            if (camPositions.LoadCamPosition("ClothingPresets") == null || force)
            {
                camPositions.StoreCamPosition("ClothingPreset");
                waitingOnScreenShot = true;
                SuperController.singleton.DoSaveScreenshot("Custom/PluginData/Chokaphi/PresetMaker/ClothingPresets", RecordCamPos);
                help.text = "Set Camera you want to use for Clothing Preset";
                help.color = Color.yellow;

                yield return new WaitWhile(() => waitingOnScreenShot);
                yield return new WaitForSecondsRealtime(1);
            }

            if (camPositions.LoadCamPosition("SkinPresets") == null || force)
            {
                camPositions.StoreCamPosition("SkinPresets");
                waitingOnScreenShot = true;
                SuperController.singleton.DoSaveScreenshot("Custom/PluginData/Chokaphi/PresetMaker/SkinPresets", RecordCamPos);
                help.text = "Set Camera you want to use for Skin Preset";
                help.color = Color.red;

                yield return new WaitWhile(() => waitingOnScreenShot);
                yield return new WaitForSecondsRealtime(1);
            }

            if (camPositions.LoadCamPosition("HairPresets") == null || force)
            {
                camPositions.StoreCamPosition("HairPresets");
                waitingOnScreenShot = true;
                SuperController.singleton.DoSaveScreenshot("Custom/PluginData/Chokaphi/PresetMaker/HairPresets", RecordCamPos);
                help.text = "Set Camera you want to use for Hair Preset";
                help.color = Color.green;

                yield return new WaitWhile(() => waitingOnScreenShot);
                yield return new WaitForSecondsRealtime(1);
            }

            help.text = camPositions.OrigHelpText;
            help.color = camPositions.OrigHelpColor;

            FreeControllerV3 headController = containingAtom.freeControllers.Where(x => x.name == "headControl").FirstOrDefault();
            SuperController.singleton.FocusOnController(headController, true, true);
        }

        void RecordCamPos(string presetType)
        {
            presetType = FileManagerSecure.GetFileName(presetType);
            camPositions.StoreCamPosition(presetType);
            waitingOnScreenShot = false;
        }

        void SavePreset(string presetType = "ClothingPresets")
        {
            //turn off auto expressions,blink, jaw collision
            containingAtom.GetStorableByID("EyelidControl").SetBoolParamValue("blinkEnabled", false);
            containingAtom.GetStorableByID("AutoExpressions").SetBoolParamValue("enabled", false);
            containingAtom.GetStorableByID("TongueControl").SetBoolParamValue("tongueCollision", false);

            List<MeshVR.PresetManagerControl> pmcs = containingAtom.presetManagerControls;
            MeshVR.PresetManagerControl presetControl = pmcs.Where(x => x.name == presetType).FirstOrDefault();
            //set preset name
            presetControl.SetStringParamValue("presetName", filePaths.StringParamValue);
            //store preset
            presetControl.CallAction("StorePreset");
            //now call the screenshot with callback
            waitingOnScreenShot = true;
        }

        void ScreenShotFinished(string text)
        {
            waitingOnScreenShot = false;
        }
        #region Save/Load
        //load cams from scene
        public override void LateRestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, bool setMissingToDefault = true)
        {
            base.LateRestoreFromJSON(jc, restorePhysical, restoreAppearance, setMissingToDefault);
            try
            {
                if (jc["AppearancePresets"] != null)
                {
                    camPositions.StoreCamPosition("AppearancePresets").RestoreFromJSON(jc["AppearancePresets"].AsObject);
                }
                if (jc["ClothingPresets"] != null)
                {
                    camPositions.StoreCamPosition("ClothingPresets").RestoreFromJSON(jc["ClothingPresets"].AsObject);
                }
                if (jc["HairPresets"] != null)
                {
                    camPositions.StoreCamPosition("HairPresets").RestoreFromJSON(jc["HairPresets"].AsObject);
                }
                if (jc["SkinPresets"] != null)
                {
                    camPositions.StoreCamPosition("SkinPresets").RestoreFromJSON(jc["SkinPresets"].AsObject);
                }

                if (jc["HeadPosition"] != null)
                {
                    //get character head position
                    FreeControllerV3 headController = containingAtom.freeControllers.Where(x => x.name == "headControl").FirstOrDefault();
                    SuperController.singleton.FocusOnController(headController, true, true);
                    camPositions.StoreCamPosition("HeadPosition", _headTransform).RestoreFromJSON(jc["HeadPosition"].AsObject);
                }
            }
            catch (Exception e)
            {
                SuperController.LogError("PresetMaker: Error Loading save.\n" + e.Message);
            }
        }

        //save cams to scene
        public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false)
        {
            JSONClass jc = base.GetJSON(includePhysical, includeAppearance, true);

            try
            {
                jc["AppearancePresets"] = camPositions.LoadCamPosition("AppearancePresets").GetJSON();
                jc["ClothingPresets"] = camPositions.LoadCamPosition("ClothingPresets").GetJSON();
                jc["HairPresets"] = camPositions.LoadCamPosition("HairPresets").GetJSON();
                jc["SkinPresets"] = camPositions.LoadCamPosition("SkinPresets").GetJSON();
                jc["HeadPosition"] = camPositions.LoadCamPosition("HeadPosition").GetJSON();
            }
            catch(Exception e) 
            {
                SuperController.LogError("PresetMaker: Error while saving.\n" + e.Message);
            }

            return jc;
        }


        #endregion


        #region ProcessLooks helpers
        JSONNode LoadJsonString(string filePath)
        {
            try
            {
                string jsonString = FileManagerSecure.ReadAllText(filePath);
                //replace all SELF with var path just to load the person node with internal VAR ref
                string varName = FileManagerSecure.NormalizePath(filePath);
                varName = Regex.Replace(varName, "(.*)(:.+)", "$1");
                jsonString = Regex.Replace(jsonString, "(SELF)(:.+)", varName + "$2");

                return JSON.Parse(jsonString);
            }
            catch (Exception e)
            {
                SuperController.LogMessage("PresetMaker: Unable to load JSON File at " + filePath);
                SuperController.LogError(e.Message);
                return new JSONNode();
            }
        }

        void LoadPersonNode(JSONNode personNode)
        {
            JSONClass asObject = personNode.AsObject;
            //load person file
            containingAtom.PreRestore(false, true);
            containingAtom.Restore(asObject, false, true, false, null, false, false, true, false);
            containingAtom.LateRestore(asObject, false, true, false, false, true, false);
            containingAtom.PostRestore(false, true);

            if (SuperController.singleton != null && asObject["id"] != null)
            {
                SuperController.singleton.RenameAtom(containingAtom, asObject["id"]);
                containingAtom.ResetPhysics(false, true);
            }
        }
        

        bool filter(string filePath)
        {
            string creatorName = Regex.Replace(filePath, @".*\\(.*)(\..*\..*\..*:)(.+)", "$1");

            if (jssc.val == "All")
                return false;

            if (creatorName == jssc.val)
                return false;

            return true;
        }
        
        void StopProcessing(bool forced = false)
        {
            if (forced && IsProcessing)
            {
                infoBox.Message = "Processing stopped by user.";
            }
            else
            {
                infoBox.Message = "Finished Proccessing all VAR json files.\nFiles processed = " + fileCount + "  looks processed = " + lookCount;
            }
            EnableUIButtons(true);
            StopRequested = false;
            IsProcessing = false;
            //reset cam to default.
            camPositions.LoadCamPosition("VAMCAM");
          
            FreeControllerV3 headController = containingAtom.freeControllers.Where(x => x.name == "headControl").FirstOrDefault();
            SuperController.singleton.FocusOnController(headController, true, true);
        }
        #endregion

        IEnumerator ProcessLooks(string path)
        {
            fileCount = 0;
            lookCount = 0;
            IsProcessing = true;
            EnableUIButtons(false);

            if (path == "AddonPackages")
            {
                if (jssc.val == "All")
                {
                    foreach (string filePath in GetFiles(path, "*.json"))
                    {
                        if (StopRequested)
                        {
                            StopProcessing(true);
                            yield break;
                        }
                        yield return StartCoroutine(ProcessFile(filePath));
                        yield return new WaitForEndOfFrame();
                    }
                }
                else if (CreatorPaths[jssc.val] != null && CreatorPaths[jssc.val].Count > 0)
                {
                    foreach (string varPath in CreatorPaths[jssc.val])
                    {
                        foreach (string filePath in GetFiles(varPath + ":", "*.json"))
                        {
                            if (StopRequested)
                            {
                                StopProcessing(true);
                                yield break;
                            }
                            yield return StartCoroutine(ProcessFile(filePath));
                            yield return new WaitForEndOfFrame();
                        }
                    }
                }
            }
            else //if (path == "Saves/Person/appearance/")
            {
                foreach (string filePath in GetFiles(path, "*.json"))
                {
                    if (StopRequested)
                    {
                        StopProcessing(true);
                        yield break;
                    }
                    yield return StartCoroutine(ProcessFile(filePath));
                    yield return new WaitForEndOfFrame();
                }
            }

            yield return new WaitForEndOfFrame();

            StopProcessing();

            yield return null;
        }


        IEnumerator ProcessFile(string filePath)
        {
            infoBox.Message = "Scanning for unprocessed files " + filePath;
            yield return null;
            fileCount++;
            if (!cacheSystem.FileInCache(filePath))
            {
                JSONNode jn = LoadJsonString(filePath);

                if (jn["atoms"] != null)
                {
                   
                    JSONNode[] personNodes = jn["atoms"].Childs.Where(x => x["type"].Value == "Person").ToArray();

                    int nodeIndex = 0;
                    foreach (JSONNode personNode in personNodes)
                    {
                        infoBox.Message ="Currently processing file:\n" + FileManagerSecure.GetFileName(filePath) + "\n\nPerson Node" + nodeIndex;
                        yield return new WaitForEndOfFrame();
                        //person node has lets us skip loading duplicate figures
                        if (!cacheSystem.PersonNodeInCache(personNode, filePath))
                        {
                            lookCount++;

                            LoadPersonNode(personNode);

                            DAZCharacterRun dazCharacterRun = _dazCharacterSelector.GetComponent<DAZCharacterRun>();
                            DAZCharacter _dazCharacter = _dazCharacterSelector.selectedCharacter;

                            yield return new WaitUntil(() => _dazCharacter.ready);
                            //wait for physics to finish
                            yield return new WaitWhile(() => dazCharacterRun.IsSimulationResetting());
                            //wait for clothing and any texture loads are finished
                            yield return new WaitWhile(() => SuperController.singleton.loadingIcon.gameObject.activeSelf);
                            //wait 2 seconds for physics to settle
                            yield return new WaitForSeconds(2);

                            //ZERO low value morphs micro values in on versions will cause duplicates any value < 0.01
                            try
                            {
                                if (_dazCharacterSelector.morphsControlFemaleUI.enabled)
                                {
                                    _dazCharacterSelector.morphsControlFemaleUI.ZeroNearZero();
                                }
                                if (_dazCharacterSelector.morphsControlMaleUI.enabled)
                                {
                                    _dazCharacterSelector.morphsControlMaleUI.ZeroNearZero();
                                }
                            }
                            catch
                            {
                            }

                            //for the first screenshot of session allow user to position camera for screenshots
                            if (needToSaveCams)
                            {
                                yield return StartCoroutine(StoreCameraPostions());
                            }

                            string varName = FileManagerSecure.NormalizePath(filePath);
                            varName = Regex.Replace(varName, "(.*)(:.+)", "$1");

                            filePaths = new FilePaths();
                            filePaths.CreateFileName(varName, personNode["id"], _dazCharacter.isMale);
                                
                            //now store screenshots based on selections
                            foreach (JSONStorableBool jsonBool in _saveOptions)
                            {
                                //if we have a match in cache skip
                                if (jsonBool.val) // && !cacheSystem.StorableInCache(filePath, personNode, personNodes.Length, jsonBool.name, value))
                                {
                                    SavePreset(jsonBool.name);
                                    //store a hash of the saved VAP file. Our storables routine doesnt return the same order as VAM's

                                    if (!cacheSystem.StorableInCache(filePath, personNode, jsonBool.name, filePaths.GetPresetPath(jsonBool.name)))
                                    {
                                        //now store the correct CAM Screenshot
                                        SuperController.singleton.DoSaveScreenshot(filePaths.GetScreenShotPath(jsonBool.name), ScreenShotFinished);

                                        //adjust cam position based on Character height  
                                        camPositions.UpdateHeadDeltas(_headTransform);
                                        camPositions.LoadCamPosition(jsonBool.name);


                                      //take two frames for camera to reposition.
                                        yield return new WaitForEndOfFrame();
                                        yield return new WaitForEndOfFrame();
                                        //yield return new WaitUntil(() => Input.GetKey(KeyCode.DownArrow));

                                        SuperController.singleton.SetLeftSelect();
                                        yield return new WaitWhile(() => waitingOnScreenShot);
                                    }
                                    else
                                    {
                                        //if already exists delete created VAP
                                        List<PresetManagerControl> pmcs = containingAtom.presetManagerControls;
                                        PresetManagerControl presetManagerControl = pmcs.Where(x => x.name == jsonBool.name).FirstOrDefault();
                                        PresetManager presetManager = presetManagerControl.GetComponent<PresetManager>();
                                        presetManager.DeletePreset();
                                    }
                                }

                            }
                            //reset cam for next load. Not needed but makes it look better if watching
                            camPositions.LoadCamPosition("VAMCAM");

                            nodeIndex++;
                        }
                    }
                }
            }
            
        }

        void Update()
        {
            if (!StopRequested && Input.GetKeyDown("space"))
            {
                StopRequested = true;
            }
        }

        static IEnumerable<string> GetFiles(string path, string pattern)
        {
            Queue<string> queue = new Queue<string>();
            queue.Enqueue(path);
            while (queue.Count > 0)
            {
                path = queue.Dequeue();
                try
                {
                    foreach (string subDir in FileManagerSecure.GetDirectories(path))
                    {
                        queue.Enqueue(subDir);
                    }
                }
                catch (Exception ex)
                {
                    SuperController.LogError(ex.Message);
                }
                string[] files = null;
                try
                {
                    files = FileManagerSecure.GetFiles(path, pattern);
                }
                catch (Exception ex)
                {
                    SuperController.LogError(ex.Message);
                }
                if (files != null)
                {
                    for (int i = 0; i < files.Length; i++)
                    {
                        yield return files[i];
                    }
                }
            }
        }


        #region


        #endregion

    }

    class FilePaths
    {
        //the name and path to save the preset file. VAM will add 'Preset_' to this for saved VAP file and save in correct folder
        //ex /PresetMaker/MYNAME  becomes  "Custom/Atom/Person/Appearance/PresetMaker/Preset_MYNAME.vap
        public string StringParamValue { get { return presetSubPath + generatedName; } }
        string presetSubPath { get; set; }
        string generatedName { get; set; }
        public string fullVarName { get; private set; }
        public string varName { get; private set; }
        public string authorName { get; private set; }
        

        string appearancePath = "Custom/Atom/Person/Appearance";
        string clothingPath = "Custom/Atom/Person/Clothing";
        string skinPath = "Custom/Atom/Person/Skin";
        string hairPath = "Custom/Atom/Person/Hair";

        public string GetScreenShotPath(string presetType)
        {
            return PresetTypePath(presetType) + presetSubPath + "Preset_" + generatedName + ".jpg";
        }

        public string GetPresetPath(string presetType)
        {
            return PresetTypePath(presetType) + presetSubPath + "Preset_" + generatedName + ".vap";
        }

        string PresetTypePath(string presetType)
        {
            switch (presetType)
            {
                case "AppearancePresets":
                    return "Custom/Atom/Person/Appearance";

                case "ClothingPresets":
                    return "Custom/Atom/Person/Clothing";

                case "SkinPresets":
                    return "Custom/Atom/Person/Skin";

                case "HairPresets":
                    return "Custom/Atom/Person/Hair";
            }

            return null;
        }

        public void CreateFileName(string fullVarName, string name = "Person", bool isMale = false)
        {
            string sex = "Female";
            if (isMale)
                sex = "Male";

            this.fullVarName = fullVarName;
            string[] varNames = fullVarName.Split('.');

            authorName = varNames[0];
            varName = varNames[1];

            presetSubPath = string.Concat(new string[]
                {
                    "/",
                    "PresetMaker",

                    "/",
                    authorName,
                    "/",
                });

            if (name.ToLower().Contains("person"))
            {
                name = varName;
            }

            int count = 0;

            string baseName = name + "-" + sex;
            generatedName = baseName;
            string path = presetSubPath + "Preset_" + baseName + ".vap";
            while (FileExists(path))
            {
                count++;
                generatedName = baseName + count;
                path = presetSubPath + "Preset_" + generatedName + ".vap";
            }
        }

        bool FileExists(string path)
        {
            if (FileManagerSecure.FileExists(appearancePath + path) || FileManagerSecure.FileExists(clothingPath + path) || FileManagerSecure.FileExists(skinPath + path) || FileManagerSecure.FileExists(hairPath + path))
            {
                return true;

            }
            return false;
        }
    }

    class CacheSystem
    {
        JSONClass cacheFileJSON = new JSONClass();
        readonly string Cachepath = "Custom/PluginData/Chokaphi/PresetMaker/Cache.json";
        MVRScript mVR;

        public CacheSystem(MVRScript mVR)
        {
            this.mVR = mVR;
            LoadCache();
        }

        public bool FileInCache(string filePath)
        {
            filePath = FileManagerSecure.NormalizePath(filePath);
            if (cacheFileJSON[filePath] == null)
            {
                return false;
            }

           foreach (JSONNode pNode in  cacheFileJSON[filePath].Childs)
            {
                if (pNode.Childs.Count() == 4)
                {
                    //do nothing
                }
                else
                {
                    return false;
                } 
            }

            return true;
        }

        //Used to remove Animations patterns from personNode. Allowing fast hash calculation for duplicate check
        string CleanPersonNodeHash(JSONNode personNode)
        {
            List<JSONNode> animationNodes = new List<JSONNode>();
            foreach (JSONNode storable in personNode["storables"].AsArray)
            {
                if (storable["id"].Value.Contains("Animation"))
                {
                    animationNodes.Add(storable);
                }
            }
            foreach (JSONNode animNode in animationNodes)
            {
                personNode["storables"].AsArray.Remove(animNode);
            }
            return ComputeNodeHash(personNode);
        }

        public bool PersonNodeInCache(JSONNode personNode, string filePath)
        {
            filePath = FileManagerSecure.NormalizePath(filePath);
            string hash = CleanPersonNodeHash(personNode);

            if (cacheFileJSON[filePath][hash] != null)
            {
                if (cacheFileJSON[filePath][hash].Childs.Count() == 4)
                {
                    //SuperController.LogError("PersonNode has 4 entries. Skipping node " + filePath);
                    return true;
                }
            }
           
            return false;
        }

        public bool StorableInCache(string filePath, JSONNode personNode, string presetType, string vapFilePath)
        {
            try
            {
                filePath = FileManagerSecure.NormalizePath(filePath);
                string vapFile = FileManagerSecure.ReadAllText(vapFilePath);

                string cacheSig = ComputeNodeHash(vapFile);
                string personNodeHash = ComputeNodeHash(personNode);

                //if we have a matching personNode and presetType already return true
                if (cacheFileJSON[filePath][personNodeHash][presetType] != null && cacheFileJSON[filePath][personNodeHash][presetType].Value == cacheSig)
                {
                    return true;
                }

                //create cache entry
                bool result = CreateNodeCacheEntry(filePath, personNodeHash, presetType, cacheSig);

                SaveCache();
                return result;
            }
            catch (Exception e)
            {
                SuperController.LogMessage("Error in  StorableInCache routine. " + e.ToString());
                return true;
            }

        }

        public bool InNodeCache(string cacheSig, string filePath)
        {
            filePath = FileManagerSecure.NormalizePath(filePath);
            if (cacheFileJSON["nodeHashes"][cacheSig] != null)
            {
                return true;
            }
            cacheFileJSON["nodeHashes"][cacheSig] = filePath;
            return false;
        }

        bool CreateNodeCacheEntry(string filePath,string personNodeHash, string presetType, string cacheSig)
        {
            if (cacheFileJSON[filePath] == null)
            {
                cacheFileJSON[filePath] = new JSONClass();
            }

            if (cacheFileJSON[filePath][personNodeHash] == null)
            {
                cacheFileJSON[filePath][personNodeHash] = new JSONClass();
            }

            if (cacheFileJSON[filePath][personNodeHash][presetType] == null)
            {
                if (InNodeCache(cacheSig, filePath))
                {
                    cacheFileJSON[filePath][personNodeHash][presetType] = cacheSig;
                    //cacheFileJSON[filePath][personNodeHash][presetType + "Duplicate"].AsBool = true;
                    return true;
                }
                else
                {
                    cacheFileJSON[filePath][personNodeHash][presetType] = cacheSig;
                }
            }

            return false;
        }

        void LoadCache()
        {
            try
            {
                string jsonString = FileManagerSecure.ReadAllText(Cachepath);
                cacheFileJSON = (JSONClass)JSON.Parse(jsonString);
                UpGradeCache();

            }
            catch (Exception e)
            {
                SuperController.LogMessage("Cache not found or was unreadable. New Cache created.");
                cacheFileJSON = new JSONClass();
                UpGradeCache();
            }

            SaveCache();
        }

        public void SaveCache()
        {
            try
            {
                mVR.SaveJSON(cacheFileJSON, Cachepath);
            }
            catch (Exception e)
            {
                SuperController.LogMessage("Failed to save Cache " + e);
            }
        }

        void UpGradeCache()
        {
            if (!cacheFileJSON.HasKey("version") || cacheFileJSON["version"].Value == "1")
            {
                SuperController.LogMessage("New cache created. This will cause duplicate entries for any previously created presets.");
                cacheFileJSON = new JSONClass();
                cacheFileJSON["version"].AsInt = 2;
                cacheFileJSON["nodeHashes"] = new JSONClass();
            }
        }

        #region Hash Functions

        public string ComputeNodeHash(JSONNode node)
        {
            if (node == null)
                return null;

            string hash = ComputeNodeHash(node.ToString());
            return hash;
        }

        public string ComputeNodeHash(string text)
        {
            if (string.IsNullOrEmpty(text))
                return null;

            byte[] raw = System.Text.Encoding.ASCII.GetBytes(text);
            return ComputeNodeHash(raw);
        }

        string ComputeNodeHash(byte[] inputBytes)
        {
            if (inputBytes == null || inputBytes.Length == 0)
                return null;

            using (MD5 md5 = MD5.Create())
            {
                byte[] bytes = md5.ComputeHash(inputBytes);
                StringBuilder builder = new StringBuilder(bytes.Length * 2);
                for (int i = 0; i < bytes.Length; i++)
                {
                    builder.Append(bytes[i].ToString("x2"));
                }
                return builder.ToString();
            }
        }
        #endregion
    }

    class TransformData
    {
        public Vector3 localPosition { get; set; }
        public Vector3 position { get; private set; }
        public Quaternion localRotation { get; private set; }
        public Vector3 centerCamEulerAngles { get; private set; }
        public float playerHeightAdjust { get; set; }

        public Color OrigHelpColor { get; set; }
        public string OrigHelpText { get; set; }

        public TransformData()
        {
            localPosition = SuperController.singleton.navigationRig.transform.localPosition;
            localRotation = SuperController.singleton.navigationRig.transform.localRotation;
            position = SuperController.singleton.navigationRig.transform.position;

            playerHeightAdjust = SuperController.singleton.playerHeightAdjust;
            centerCamEulerAngles = SuperController.singleton.MonitorCenterCamera.transform.localEulerAngles;
        }

        public TransformData(Transform transform)
        {   //used for head delta and world position
            position = transform.position;
            localRotation = transform.rotation;
            playerHeightAdjust = SuperController.singleton.playerHeightAdjust;
            centerCamEulerAngles = SuperController.singleton.MonitorCenterCamera.transform.localEulerAngles;
        }

        public void RestoreTransform(Vector3 delta = new Vector3())
        {
            if (delta == null)
                delta = Vector3.zero;

            //Vector3 newPos = new Vector3(localPosition.x + delta.x, localPosition.y, localPosition.z + delta.z);
            SuperController.singleton.navigationRig.transform.localPosition = localPosition;
            SuperController.singleton.navigationRig.transform.localRotation = localRotation;
            SuperController.singleton.playerHeightAdjust = playerHeightAdjust + delta.y;
            SuperController.singleton.MonitorCenterCamera.transform.localEulerAngles = centerCamEulerAngles;
        }

        #region save-load
        //convert TransforData to JSON Object for storage
        public JSONClass GetJSON()
        {
            JSONClass jc = new JSONClass();

            if (localPosition == null || localRotation == null || centerCamEulerAngles == null)
                return null;

            try
            {
                jc["localPosition"]["x"].AsFloat = localPosition.x;
                jc["localPosition"]["y"].AsFloat = localPosition.y;
                jc["localPosition"]["z"].AsFloat = localPosition.z;

                jc["Rotation"]["x"].AsFloat = localRotation.x;
                jc["Rotation"]["y"].AsFloat = localRotation.y;
                jc["Rotation"]["z"].AsFloat = localRotation.z;
                jc["Rotation"]["w"].AsFloat = localRotation.w;

                jc["centerCamEulerAngles"]["x"].AsFloat = centerCamEulerAngles.x;
                jc["centerCamEulerAngles"]["y"].AsFloat = centerCamEulerAngles.y;
                jc["centerCamEulerAngles"]["z"].AsFloat = centerCamEulerAngles.z;

                jc["playerHeightAdjust"].AsFloat = playerHeightAdjust;
            }
            catch (Exception e)
            {
                SuperController.LogError("PresetMaker: Error Saving JSON\n" + e.Message);
            }
            return jc;
        }

        public void RestoreFromJSON(JSONClass jc)
        {
            try
            {
                localPosition = new Vector3(jc["localPosition"]["x"].AsFloat, jc["localPosition"]["y"].AsFloat, jc["localPosition"]["z"].AsFloat);
                localRotation = new Quaternion(jc["Rotation"]["x"].AsFloat, jc["Rotation"]["y"].AsFloat, jc["Rotation"]["z"].AsFloat, jc["Rotation"]["w"].AsFloat);

                centerCamEulerAngles = new Vector3(jc["centerCamEulerAngles"]["x"].AsFloat, jc["centerCamEulerAngles"]["y"].AsFloat, jc["centerCamEulerAngles"]["z"].AsFloat);
                playerHeightAdjust = jc["playerHeightAdjust"].AsFloat;
            }
            catch (Exception e)
            {
                SuperController.LogError("PresetMaker: Error loading JSON\n" + e.Message);
            }
        }
        #endregion
    }

    class CamPositions
    {
        public string OrigHelpText { get; set; }
        public Color OrigHelpColor { get; set; }

        Vector3 deltaP = new Vector3();
        Dictionary<string, TransformData> camDictionary = new Dictionary<string, TransformData>();

        public void UpdateHeadDeltas(Transform transform)
        {
            deltaP = transform.position - camDictionary["HeadPosition"].position;
        }

        public TransformData LoadCamPosition(string presetType)
        {
            TransformData camData;
            if (camDictionary.TryGetValue(presetType, out camData))
            {
                camData.RestoreTransform(deltaP);
                return camData;
            }
            else
            {
                return null;
            }
        }

        public TransformData StoreCamPosition(string presetType, Transform transform)
        {
            return camDictionary[presetType] = new TransformData(transform);
        }

        public TransformData StoreCamPosition(string presetType)
        {
           return  camDictionary[presetType] = new TransformData();
        }

    }

    public class InfoBox
    {
        JSONStorableString infoBox;
        int varCount = 0;
        int authorCount = 0;
        string status = "Waiting on Scan to Finishe before creating UI";
        string vapScan = "";
        int vapCount = 0;
        string message = "";

        public int VarCount { set { varCount = value; Update();  } }
        public int AuthorCount { set { authorCount = value; Update(); } }
        public string Status { set { status = value; Update(); } }
        public string VapScan { set { vapScan = value; Update(); } }
        public int VapCount { set { vapCount = value; Update(); } }
        public string Message { set { message = value; Update(); } }

        public InfoBox(JSONStorableString uiInfoBox)
        {
            infoBox = uiInfoBox;
        }

        public void Update()
        {
            infoBox.SetVal( String.Format(formated, status, varCount, authorCount, vapScan, vapCount, message));
            
        }

        string formated =
@"{0}                                   {3}
Json's found {1}                                Total existing presets {4}                
Authors found {2}

{5}
        ";


    }


}
